## File: spam_filter.R
## Author: John Rothfels (rothfels)
## Edited by: Kevin Shin (hgkshin)
## Description: Naive Bayes Spam Filter Implementation
## Lecture 5 Part 1
## ---------------------------------------------------

library(kernlab)
data(spam)

RestructureSpamDataset <- function(spam) {
  ## Restures the 'spam' data frame according to the specifications in the
  ## handout (and summarized below).
  ##
  ## Args:
  ##   spam - the 'spam' data frame from the 'kernlab' package
  ## Returns:
  ##   A data frame with the first 48 columns of spam factored into
  ##   binary random variables (representing whether each word exists
  ##   or not) as well as the last column of 'spam'. The labels of the
  ##   factor for each of the 48 variables is "yes" for "it exists" and
  ##   "no" for "the word does not exist"

}

ChooseTrainingAndTesting <- function(dataset, k) {
  ## Chooses a random subset of the dataset to use for training, leaving
  ## the rest for testing.
  ##
  ## Args:
  ##   dataset - a data frame to split into training and testing data frames
  ##   k - the number of rows to include in the training data set
  ## Returns:
  ##   A list with elements $training and $testing, each a data frame with
  ##   a disjoint subset of rows taken randomly from 'dataset'.

}

GivenNonSpam <- function(training) {
  ## Computes conditional probabilities of the form P(Xi | type = "nonspam").
  ##
  ## Args:
  ##   training - data frame with training data.
  ## Returns:
  ##   A matrix with 2 rows and 48 columns, one per observed variable
  ##   from the training data frame. The first row gives the conditional
  ##   probabilities P(Xi = "no" | type = "nonspam") for each of the 48
  ##   observed variables. The second row gives conditional probabilities
  ##   P(Xi = "yes" | type = "nonspam"). Note that all columns should sum
  ##   to one.
  ## Details:
  ##   This function uses Laplace smoothing for conditional probabilities.
  ##   You can implement Laplace smoothing by imagining, for each of
  ##   the 48 columns, that we see two additional datapoints ("no" and "yes").

  ## Choose the subset of the data frame which contains only "nonspam"
  ## rows, and keep only the first 48 columns.
  subset <- NA

  ## Compute, for each of the 48 columns in the subset above, the probability
  ## of the value being "no". Remember to use Laplace smoothing.
  pNo <- NA

  return(rbind(no = pNo, yes = 1 - pNo))
}

GivenSpam <- function(training) {
  ## Computes conditional probabilities of the form P(Xi | type = "spam").
  ## Details the same as above.

  ## Choose the subset of the data frame which contains only "nonspam"
  ## rows, and keep only the first 48 columns.
  subset <- NA

  ## Compute, for each of the 48 columns in the subset above, the probability
  ## of the value being "no". Remember to use Laplace smoothing.
  pNo <- NA

  return(rbind(no = pNo, yes = 1 - pNo))
}

## In a more stylistically savvy implementation, the two functions above
## could be combined into the following:
##
##   ComputeConditionalProbabilities <- function(training, type) {
##     ## Here, 'type' is "spam" or "nonspam".
##     ...
##   }

Predict <- function(features, p.nonspam, p.spam, nonspam.probs, spam.probs) {
  ## Predicts whether the feature vector (with 48 observed variables) represents "spam"
  ##  or "nonspam" by choosing argmax P(row | type) * P(type).
  ##
  ## Args:
  ##   features - vector with 48 values (one per observed variable)
  ##   p.nonspam - P(type = "nonspam")
  ##   p.spam - P(type = "spam")
  ##   nonspam.probs - conditional probabilities of the form P(Xi | type = "nonspam")
  ##   spam.probs - conditional probabilities of the form P(Xi | type = "spam")
  ## Returns:
  ##   "spam" or "nonspam"

}

RunTrainingAndTesting <- function(k) {
  ## Runs spam filter training and testing and returns the average number of
  ## correctly classified emails.
  ## Args:
  ##   k - number of rows to use for training
  datasets <- ChooseTrainingAndTesting(RestructureSpamDataset(spam), k)
  
  log.p.nonspam <- log(mean(datasets$training$type == "nonspam"))
  log.p.spam <- log(mean(datasets$training$type == "spam"))
  log.nonspam.probs <- log(GivenNonSpam(datasets$training))
  log.spam.probs <- log(GivenSpam(datasets$training))
  responses <- sapply(1:nrow(datasets$testing),
                      function(i) Predict(unlist(datasets$testing[i, 1:48]),
                                          log.p.nonspam, log.p.spam,
                                          log.nonspam.probs, log.spam.probs))
  performance <- mean(responses == datasets$testing$type)
  return(list(responses = responses, performance = performance))
}

RunTrainingAndTesting(3000)